SystemOrganization addCategory: #'Pier-Exporter-Code'!

!Object methodsFor: '*pier-exporter-code' stamp: 'NickAger 12/12/2011 16:58'!
asCode
	^ String streamContents: [ :stream | self codeOn: stream ]! !

!Object methodsFor: '*pier-exporter-code' stamp: 'NickAger 12/12/2011 16:59'!
codeOn: aStream
	self javascriptOn: aStream! !

MAVisitor subclass: #PRCodeBuilderDescriptionVisitor
	instanceVariableNames: 'structure stream methods exporter selectorsWithCommonDefaults'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Pier-Exporter-Code'!

PRCodeBuilderDescriptionVisitor class
	instanceVariableNames: 'aComponent'!

PRCodeBuilderDescriptionVisitor class
	instanceVariableNames: 'aComponent'!

!PRCodeBuilderDescriptionVisitor class methodsFor: 'instance creation' stamp: 'NickAger 1/26/2012 17:13'!
structure: structure methods: methods exporter: exporter on: stream
	^ self basicNew
		initializeWithStructure: structure methods: methods exporter: exporter stream: stream
		yourself! !

!PRCodeBuilderDescriptionVisitor methodsFor: 'private' stamp: 'NickAger 1/26/2012 16:10'!
addCodeFor: aDescription value: value
	aDescription accessor asCodeFor: structure description: aDescription value: value on: stream.
	stream nextPut: Character cr! !

!PRCodeBuilderDescriptionVisitor methodsFor: 'modifying' stamp: 'NickAger 12/12/2011 12:29'!
addMethod: method
	self methods add: method! !

!PRCodeBuilderDescriptionVisitor methodsFor: 'initialization' stamp: 'NickAger 1/26/2012 17:14'!
initializeWithStructure: aStructure methods: methodsCollection exporter: anExporter stream: aStream
	self initialize.
	
	structure := aStructure.
	methods := methodsCollection.
	exporter := anExporter.
	stream := aStream! !

!PRCodeBuilderDescriptionVisitor methodsFor: 'check defaults' stamp: 'NickAger 1/26/2012 15:56'!
isCommonDefault: aDescription
	| accessor descriptionDefault structureValue defaultCheckBlock |
	
	"Unfortunately the description frequently doesn't define a default value, but there are a common set of defaults we don't need to write.
	We key off the readSelector and check against a dictionary for common defaults"
	accessor := aDescription accessor.
	(accessor class canUnderstand: #readSelector) ifFalse: [ ^ false ].
	defaultCheckBlock := self selectorsWithCommonDefaults at: accessor readSelector ifAbsent: [ ^ false ].
	
	^ defaultCheckBlock value! !

!PRCodeBuilderDescriptionVisitor methodsFor: 'check defaults' stamp: 'NickAger 1/26/2012 15:46'!
isDefaultDescriptionValue: aDescription
	| descriptionDefault structureValue |
	
	descriptionDefault := aDescription default.
	structureValue := aDescription accessor read: structure.
	^ descriptionDefault = structureValue ! !

!PRCodeBuilderDescriptionVisitor methodsFor: 'check defaults' stamp: 'NickAger 1/26/2012 16:36'!
isNil: aDescription
	^ (aDescription accessor read: structure) = nil
! !

!PRCodeBuilderDescriptionVisitor methodsFor: 'accessing' stamp: 'NickAger 12/12/2011 16:19'!
methods
	^ methods! !

!PRCodeBuilderDescriptionVisitor methodsFor: 'check defaults' stamp: 'NickAger 1/26/2012 15:55'!
selectorsWithCommonDefaults
	^ selectorsWithCommonDefaults ifNil: [ 
		selectorsWithCommonDefaults := Dictionary new
			at: #name put: [ true ];
			at: #title put: [ structure title = structure name capitalized ];
			at: #menu put: [ structure menu = structure title ];
			at: #tags put: [ structure tags isEmpty ];
			at: #hideFromMenus put: [ structure hideFromMenus not ];
			yourself ]! !

!PRCodeBuilderDescriptionVisitor methodsFor: 'visiting' stamp: 'NickAger 1/26/2012 16:34'!
visit: aDescription
	"only visit writable descriptions - otherwise we can't initialise it."
	(aDescription accessor canWrite: structure) ifFalse: [ ^ self ]. 
	
	"only visit descriptions with non-default values"
	(self isDefaultDescriptionValue: aDescription) ifTrue: [ ^ self ].
	(self isNil: aDescription) ifTrue: [ ^ self ].
	(self isCommonDefault: aDescription) ifTrue: [ ^ self ].

	aDescription acceptMagritte: self! !

!PRCodeBuilderDescriptionVisitor methodsFor: 'visiting' stamp: 'NickAger 12/12/2011 15:16'!
visitDescription: aDescription
	| value |
	value := (aDescription accessor read: structure) asCode.
	self addCodeFor: aDescription value: value! !

!PRCodeBuilderDescriptionVisitor methodsFor: 'visiting' stamp: 'NickAger 12/12/2011 15:15'!
visitDocumentDescription: aDescription
	"could have used:
		 value := (aDescription toString: (aDescription accessor read: structure)) asCode
	but I thought it was good to be explicit that we require wiki format output"
	| value |
	value := (PRWikiWriter write: (aDescription accessor read: structure)) asCode.
	self addCodeFor: aDescription value: value! !

!PRCodeBuilderDescriptionVisitor methodsFor: 'visiting' stamp: 'NickAger 1/26/2012 16:11'!
visitFileDescription: aDescription
	| fileModel |
	fileModel := aDescription accessor read: structure.
	stream 
		<< '	structure file: (' << fileModel class greaseString << ' new
		filename: ' << fileModel filename asCode << ';
		mimetype: ' << fileModel mimetype asCode << ';
		contents: ' << (fileModel isText ifTrue: [ fileModel contents asString asCode ] ifFalse: [ fileModel contents greaseString ]) << ').' << String cr

! !

!PRCodeBuilderDescriptionVisitor methodsFor: 'visiting' stamp: 'NickAger 1/26/2012 17:10'!
visitStructureDescription: aDescription
	| structureSelector subStructure |
	subStructure := aDescription accessor read: structure.
	aDescription accessor asCodeFor: structure description: aDescription value: ' self ', subStructure asSelector on: stream.
	stream nextPut: Character cr.
	
	self methods addAll: (exporter writeCodeFor: subStructure)! !

!Symbol methodsFor: '*pier-exporter-code' stamp: 'NickAger 12/12/2011 18:20'!
codeOn: aStream
	aStream << '#' << self! !

!MAAccessor methodsFor: '*pier-exporter-code' stamp: 'NickAger 12/11/2011 22:52'!
asCodeFor: structure description: description value: value on: stream
	| structureClass |
	structureClass := structure class greaseString.
	stream << '	' << structureClass << ' description detect: [ :each | each label = ''' << description label << ''' ]) accessor write: ' << value << ' to: structure. '! !

!PRDocumentAccessor methodsFor: '*pier-exporter-code' stamp: 'NickAger 12/11/2011 22:49'!
asCodeFor: structure description: description value: value on: stream
	stream << '	structure contents: ' << value << '.'! !

!PRStructure methodsFor: '*pier-exporter-code' stamp: 'NickAger 1/26/2012 17:05'!
asSelector
	| aName selector |
	aName := self absolutePath.
	aName := (aName = '/') ifTrue: [ self name ] ifFalse: [ aName ].

	selector := aName select: [ :each | each isAlphaNumeric ].
	[ selector first isDigit ]
		whileTrue: [ selector := selector allButFirst ].
		
	selector at: 1 put: (selector at: 1) asLowercase.
		
	^ selector asSymbol! !

!PRStructure methodsFor: '*pier-exporter-code' stamp: 'NickAger 1/26/2012 17:03'!
codeOn: aStream
	aStream << '(self ' << self asSelector << ')'! !

GRObject subclass: #PRPier2KernelCodeExporter
	instanceVariableNames: 'selectorsWithCommonDefaults exportedMethods recursiveMethods prcomponentWriter'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Pier-Exporter-Code'!

!PRPier2KernelCodeExporter class methodsFor: 'instance creation' stamp: 'NickAger 12/12/2011 22:49'!
exportKernel: aKernel
	^ self new
		exportKernel: aKernel;
		yourself! !

!PRPier2KernelCodeExporter class methodsFor: 'instance creation' stamp: 'NickAger 1/26/2012 21:07'!
exportMigratedKernel: aKernel
	^ self new
		setPrototypePRComponentGeneration;
		exportKernel: aKernel;
		yourself! !

!PRPier2KernelCodeExporter methodsFor: 'exporting' stamp: 'NickAger 12/12/2011 20:12'!
addKernelCreatorMethod: aKernel forComponent: component
	| method |
	method := String streamContents: [ :stream |
		stream << 'createKernel' << String cr
		 << '	^ PRKernel named: ''' << aKernel name << ''' root: self rootPage' ].
	self compileMethod: method intoComponent: component! !

!PRPier2KernelCodeExporter methodsFor: 'exporting' stamp: 'NickAger 1/26/2012 14:48'!
compile: aString into: aClass classified: classificationSymbol
	GRPlatform current 
		compile: aString
		into: aClass
		classified: classificationSymbol
! !

!PRPier2KernelCodeExporter methodsFor: 'exporting' stamp: 'NickAger 1/26/2012 14:49'!
compileMethod: method intoComponent: component
	self compile: method into: component classified: #'exported code'



! !

!PRPier2KernelCodeExporter methodsFor: 'exporting' stamp: 'NickAger 12/12/2011 20:12'!
compileMethods: methods intoComponent: component
	methods do: [ :method |
		self compileMethod: method intoComponent: component ]


! !

!PRPier2KernelCodeExporter methodsFor: 'exporting' stamp: 'NickAger 12/12/2011 21:28'!
createClassFor: kernel 
	^ self createClassFor: kernel instanceVariableNames: ''! !

!PRPier2KernelCodeExporter methodsFor: 'exporting' stamp: 'NickAger 1/26/2012 14:46'!
createClassFor: kernel instanceVariableNames: instanceVariableNames
	^ Object
		subclass: (self selectorFromName: kernel name, 'KernelCreator') capitalized  
		instanceVariableNames: instanceVariableNames
		classVariableNames: 'rootPage'
		poolDictionaries: ''
		category: 'Pier-Exported-Code'! !

!PRPier2KernelCodeExporter methodsFor: 'exporting' stamp: 'NickAger 1/26/2012 14:50'!
exportKernel: kernel
	| component rootComponent |
	component := self createClassFor: kernel.
	self addKernelCreatorMethod: kernel forComponent: component.
	
	rootComponent := kernel root.
	
	self compile: (self rootPageTemplate: rootComponent)
		into: component
		classified: #'exported code'.
		
	rootComponent enumerator with all do: [ :structure |
		| methods |
		methods := self writeCodeFor: structure.
		self compileMethods: methods intoComponent: component ].
	
	self rewriteRecursiveMethodsFor: kernel! !

!PRPier2KernelCodeExporter methodsFor: 'helpers' stamp: 'NickAger 1/26/2012 21:01'!
handlePRComponentClassInitialisationFor: aPRComponent on: stream
	aPRComponent settings associationsDo: [ :aDescriptionValuePair |
		stream << '	structure write: ' << aDescriptionValuePair value asCode << 
			' using: ' << aPRComponent componentClass name << ' ' 
				<< (self selectorFromDescription: aDescriptionValuePair key onClass: aPRComponent componentClass)
				<< '.' << String cr.
		
		"if the value is a structure we need to make sure we avoid a recursive loop "
		(aDescriptionValuePair value isKindOf: PRStructure) ifTrue: [ 
			recursiveMethods add: aDescriptionValuePair value asSelector ] ]! !

!PRPier2KernelCodeExporter methodsFor: 'helpers' stamp: 'NickAger 1/26/2012 22:38'!
handlePRComponentPrototypeInitialisationFor: aPRComponent on: stream
	stream << '	structure prototypeInstance: (' << aPRComponent componentClass name << ' new' << String cr.
	aPRComponent settings associationsDo: [ :aDescriptionValuePair |
		| accessor |
		accessor := aDescriptionValuePair key accessor.
		(accessor class canUnderstand: #readSelector) ifTrue: [
			stream << '		' << accessor readSelector << ': ' << aDescriptionValuePair value asCode << ';' << String cr ].
		
		"if the value is a structure we need to make sure we avoid a recursive loop "
		(aDescriptionValuePair value isKindOf: PRStructure) ifTrue: [ 
			recursiveMethods add: aDescriptionValuePair value asSelector ] ].
	stream << '		yourself).' << String cr.! !

!PRPier2KernelCodeExporter methodsFor: 'initialization' stamp: 'NickAger 1/26/2012 21:01'!
initialize
 	super initialize.
	
	exportedMethods := Set new.
	recursiveMethods := Set new.
	
	prcomponentWriter := [:aPRComponent :stream | self handlePRComponentClassInitialisationFor: aPRComponent on: stream ]! !

!PRPier2KernelCodeExporter methodsFor: 'exporting' stamp: 'NickAger 1/26/2012 20:51'!
rewriteRecursiveMethodsFor: kernel
	| instanceVariables component |
	instanceVariables := String streamContents: [ :stream |
		recursiveMethods do: [ :each | 
			stream nextPutAll: each ]
			separatedBy: [ stream nextPut: Character space ] ].
		
	component := self createClassFor: kernel instanceVariableNames: instanceVariables.
	
	"this is somewhat hacky - we rewrite reentrant component creation methods with lazy initialiser using a regex match
	a two pass generator would remove the need for this hack - but would be more complex. "
	recursiveMethods do: [ :each | 
		| regex initialAssignment propertySetting method |
		regex := '.*(structure \:=[[:print:]]+)(.*)\^ structure' asRegex.
		regex search: (component sourceCodeAt: each).
		self assert: (regex subexpressionCount = 3).
		initialAssignment :=  regex subexpression: 2.
		propertySetting :=  regex subexpression: 3.
		method := String streamContents: [ :stream |
			stream 
				<<  each << String cr
				<< '	 ^ ' << each << ' ifNil: [ ' << String cr
				<< '		| structure |' << String cr
				<< '		' << initialAssignment << String cr
				<< '		' << each << ' := structure.'
				<< propertySetting << String cr		
				<< '		' << each << ' ]' ].
				
		self compileMethod: method intoComponent: component ]! !

!PRPier2KernelCodeExporter methodsFor: 'exporting' stamp: 'NickAger 1/26/2012 17:11'!
rootPageTemplate: rootComponent
	^ String streamContents: [ :stream | 
		stream << 'rootPage
	^ rootPage ifNil: [
		rootPage := self ' << (rootComponent asSelector) << ' ]']! !

!PRPier2KernelCodeExporter methodsFor: 'helpers' stamp: 'NickAger 1/26/2012 17:49'!
selectorFromDescription: aDescription onClass: aClass
	| descriptionSelectors |
	descriptionSelectors := SortedCollection withAll: (aClass class allSelectors select: [ :each | each isDescriptionDefinition ]).		
	
	^ descriptionSelectors detect: [ :each |  (aClass perform: each) = aDescription ]
	! !

!PRPier2KernelCodeExporter methodsFor: 'helpers' stamp: 'NickAger 12/12/2011 20:13'!
selectorFromName: aName
	| selector |
	selector := aName select: [ :each | each isAlphaNumeric ].
	[ selector first isDigit ]
		whileTrue: [ selector := selector allButFirst ].
		
	selector at: 1 put: (selector at: 1) asLowercase.
		
	^ selector asSymbol! !

!PRPier2KernelCodeExporter methodsFor: 'initialization' stamp: 'NickAger 1/26/2012 21:07'!
setPrototypePRComponentGeneration
	prcomponentWriter := [ :aPRComponent :stream | self handlePRComponentPrototypeInitialisationFor: aPRComponent on: stream ]! !

!PRPier2KernelCodeExporter methodsFor: 'helpers' stamp: 'NickAger 1/26/2012 22:17'!
writeCodeFor: structure
	| methods method codeBuilder selector |
	selector := structure asSelector.
	
	"recursion guard"
	(exportedMethods includes: selector)
		ifTrue: [ recursiveMethods add: selector. ^ #() ]
		ifFalse: [ exportedMethods add: selector ].
	
	methods := OrderedCollection new.
	
	method :=  String streamContents: [ :stream | 
		stream << selector << '
	| structure  |
	structure := ' << structure class greaseString << ' named: ''' << structure name << '''.' << String cr.
	
		codeBuilder := PRCodeBuilderDescriptionVisitor structure: structure methods: methods exporter: self  on: stream.
		structure description do: [ :aDescription |
			codeBuilder visit: aDescription ].
		
		structure class name = #PRComponent ifTrue: [ prcomponentWriter value: structure value: stream  ].
			
		structure enumerator do: [ :each | 
			stream << '	structure addChild: self ' << (each asSelector) << '.' << String cr ].
			
		stream << '	^ structure' ].
	
	methods add: method.
	^ methods! !

!MASelectorAccessor methodsFor: '*pier-exporter-code' stamp: 'NickAger 12/11/2011 22:52'!
asCodeFor: structure description: description value: value on: stream
	stream << '	structure ' << self writeSelector greaseString << ' ' << value << '.'! !

!UndefinedObject methodsFor: '*pier-exporter-code' stamp: 'NickAger 1/26/2012 16:18'!
codeOn: aStream
	aStream << 'nil'! !

!DateAndTime methodsFor: '*pier-exporter-code' stamp: 'NickAger 12/12/2011 23:34'!
codeOn: aStream
	aStream << '(' << self class greaseString << ' year: '; print: self year; << ' month: '; print: self month; << ' day: '; print: self dayOfMonth; << ')' ! !

!Collection methodsFor: '*pier-exporter-code' stamp: 'NickAger 12/11/2011 20:49'!
codeOn: aStream
	aStream << '#('.
	self do: [ :each |
		each codeOn: aStream ].
	aStream << ')'! !

!Array methodsFor: '*pier-exporter-code' stamp: 'NickAger 1/26/2012 22:35'!
codeOn: aStream
	aStream nextPut: ${.
	self do: [:el | el storeOn: aStream] separatedBy: [ aStream nextPutAll: '. '].
	aStream nextPut: $}! !
